import tensorflow as tf
import numpy as np
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt

class SmartVisionCNN:
    def __init__(
        self,
        input_shape=(28, 28, 1),
        num_classes=10,
        regularization=0.001,
        dropout_rate=0.5,
        learning_rate=0.001,
        loss_fn='sparse_categorical_crossentropy',
        metrics_fn=['accuracy']
    ):
        self.input_shape = input_shape
        self.num_classes = num_classes
        self.loss_fn = loss_fn
        self.metrics_fn = metrics_fn

        regularizer = tf.keras.regularizers.l2(regularization)

        self.model = tf.keras.Sequential([
            tf.keras.layers.InputLayer(input_shape=input_shape),
            tf.keras.layers.Conv2D(32, (3, 3), padding='same', kernel_regularizer=regularizer),
            tf.keras.layers.BatchNormalization(),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPooling2D(pool_size=(2, 2)),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(128, kernel_regularizer=regularizer),
            tf.keras.layers.BatchNormalization(),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Dropout(dropout_rate),
            tf.keras.layers.Dense(num_classes, activation='softmax')
        ])

        self.model.compile(
            optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate),
            loss=loss_fn,
            metrics=metrics_fn
        )

    def split_data(self, X, y, val_size=0.2, test_size=0.0, random_state=42):
        """
        Split data into train/validation(/test) sets.
        """
        if test_size > 0:
            X_temp, X_test, y_temp, y_test = train_test_split(X, y, test_size=test_size, random_state=random_state)
            X_train, X_val, y_train, y_val = train_test_split(X_temp, y_temp, test_size=val_size, random_state=random_state)
            return X_train, X_val, X_test, y_train, y_val, y_test
        else:
            X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=val_size, random_state=random_state)
            return X_train, X_val, y_train, y_val

    def train_model(self, X_train, y_train, X_val=None, y_val=None, epochs=10, batch_size=32, save_model_path='model.keras'):
        """
        Train the CNN model.
        """
        callbacks_fn = [
            tf.keras.callbacks.EarlyStopping(patience=3, monitor='val_loss', restore_best_weights=True),
            tf.keras.callbacks.ModelCheckpoint(save_model_path, save_best_only=True, monitor='val_loss'),
            tf.keras.callbacks.ReduceLROnPlateau(monitor='val_loss', patience=2, factor=0.5),
        ]

        val_data = (X_val, y_val) if X_val is not None and y_val is not None else None

        self.history = self.model.fit(
            X_train, y_train,
            validation_data=val_data,
            epochs=epochs,
            batch_size=batch_size,
            callbacks=callbacks_fn,
            verbose=1
        )

    def test_model(self, X_test, y_test):
        """
        Evaluate the model on test data.
        """
        return self.model.evaluate(X_test, y_test, verbose=1)

    def predict_model(self, X_input):
        """
        Make predictions from input data.
        """
        predictions = self.model.predict(X_input)
        return np.argmax(predictions, axis=1)

    def save_model(self, model_path='model.keras'):
        """
        Save the trained model to disk.
        """
        self.model.save(model_path)

    def load_model(self, model_path):
        """
        Load a saved model from disk.
        """
        self.model = tf.keras.models.load_model(model_path)

    def model_summary(self):
        """
        Print the summary of the CNN model.
        """
        self.model.summary()

    def plot_accuracy(self):
        """
        Plot training & validation accuracy and loss.
        """
        if not hasattr(self, 'history') or self.history is None:
            raise ValueError("You must train the model before plotting.")

        history = self.history.history

        plt.figure(figsize=(12, 5))

        # Accuracy
        plt.subplot(1, 2, 1)
        plt.plot(history.get('accuracy', []), label='Train Accuracy')
        if 'val_accuracy' in history:
            plt.plot(history['val_accuracy'], label='Validation Accuracy')
        plt.title('Model Accuracy')
        plt.xlabel('Epoch')
        plt.ylabel('Accuracy')
        plt.legend()

        # Loss
        plt.subplot(1, 2, 2)
        plt.plot(history.get('loss', []), label='Train Loss')
        if 'val_loss' in history:
            plt.plot(history['val_loss'], label='Validation Loss')
        plt.title('Model Loss')
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.legend()

        plt.tight_layout()
        plt.show()
