import tensorflow as tf
import tensorflow_hub as hub
import cv2
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import os


class SVOL:
    def __init__(self):
        self.model = hub.load("https://tfhub.dev/tensorflow/efficientdet/d0/1")
        self.class_names = self._load_coco_labels()

    def _load_coco_labels(self):
        return [
            "background", "person", "bicycle", "car", "motorcycle", "airplane", "bus",
            "train", "truck", "boat", "traffic light", "fire hydrant", "stop sign",
            "parking meter", "bench", "bird", "cat", "dog", "horse", "sheep", "cow",
            "elephant", "bear", "zebra", "giraffe", "backpack", "umbrella", "handbag",
            "tie", "suitcase", "frisbee", "skis", "snowboard", "sports ball", "kite",
            "baseball bat", "baseball glove", "skateboard", "surfboard", "tennis racket",
            "bottle", "wine glass", "cup", "fork", "knife", "spoon", "bowl", "banana",
            "apple", "sandwich", "orange", "broccoli", "carrot", "hot dog", "pizza",
            "donut", "cake", "chair", "couch", "potted plant", "bed", "dining table",
            "toilet", "tv", "laptop", "mouse", "remote", "keyboard", "cell phone",
            "microwave", "oven", "toaster", "sink", "refrigerator", "book", "clock",
            "vase", "scissors", "teddy bear", "hair drier", "toothbrush"
        ]

    def Preprocess(self, image_path):
        image = cv2.imread(image_path)
        image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image_tensor = tf.convert_to_tensor(image_rgb, dtype=tf.uint8)
        image_tensor = tf.expand_dims(image_tensor, axis=0)
        return image_tensor, image_rgb

    def Model_Predict(self, image_path, score_threshold=0.5):
        image_tensor, original_image = self.Preprocess(image_path)
        results = self.model(image_tensor)

        boxes = results['detection_boxes'].numpy()[0]
        scores = results['detection_scores'].numpy()[0]
        classes = results['detection_classes'].numpy()[0].astype(int)

        height, width, _ = original_image.shape
        output = []
        for box, score, cls in zip(boxes, scores, classes):
            if score >= score_threshold:
                ymin, xmin, ymax, xmax = box
                left, top = int(xmin * width), int(ymin * height)
                right, bottom = int(xmax * width), int(ymax * height)
                output.append({
                    'box': [left, top, right, bottom],
                    'score': float(score),
                    'class': self.class_names[cls] if cls < len(self.class_names) else f'class_{cls}'
                })

        return {'results': output, 'image': original_image}

    def Draw_Boxes(self, prediction_result):
        image = prediction_result['image']
        boxes_data = prediction_result['results']

        fig, ax = plt.subplots(1)
        ax.imshow(image)

        for item in boxes_data:
            left, top, right, bottom = item['box']
            score = item['score']
            label = item['class']

            rect = patches.Rectangle(
                (left, top), right - left, bottom - top,
                linewidth=2, edgecolor='red', facecolor='none'
            )
            ax.add_patch(rect)
            ax.text(left, top - 10, f"{label}: {score:.2f}", color='white', fontsize=8, backgroundcolor='red')

        plt.axis('off')
        plt.show()

    def Predict_From_Webcam(self, score_threshold=0.5):
        cap = cv2.VideoCapture(0)
        print("📷 Webcam is active. Press 'q' to quit.")

        while cap.isOpened():
            ret, frame = cap.read()
            if not ret:
                break

            rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            image_tensor = tf.convert_to_tensor(rgb, dtype=tf.uint8)
            image_tensor = tf.expand_dims(image_tensor, axis=0)

            results = self.model(image_tensor)
            boxes = results['detection_boxes'].numpy()[0]
            scores = results['detection_scores'].numpy()[0]
            classes = results['detection_classes'].numpy()[0].astype(int)

            height, width, _ = frame.shape
            for box, score, cls in zip(boxes, scores, classes):
                if score < score_threshold:
                    continue
                ymin, xmin, ymax, xmax = box
                start = (int(xmin * width), int(ymin * height))
                end = (int(xmax * width), int(ymax * height))
                label = self.class_names[cls] if cls < len(self.class_names) else str(cls)

                cv2.rectangle(frame, start, end, (0, 0, 255), 2)
                cv2.putText(frame, f"{label}: {score:.2f}", (start[0], start[1] - 10),
                            cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 255, 0), 2)

            cv2.imshow("SmartVision (SVOL)", frame)

            if cv2.waitKey(1) & 0xFF == ord('q'):
                break

        cap.release()
        cv2.destroyAllWindows()

    def Train_Model(self, 
                    model_name="trained_model", 
                    dataset=None, epochs=5, batch_size=32, 
                    loss='categorical_crossentropy', 
                    optimizer='adam', 
                    custom_layers=None):
        """
        Train a custom object detection model.

        Parameters:
        - model_name: name to save the model as.
        - dataset: a tuple (train_data, val_data) of tf.data.Dataset or None to use tf.keras.datasets.cifar10
        - epochs: number of training epochs.
        - batch_size: batch size for training.
        - loss: loss function.
        - optimizer: optimizer type.
        - custom_layers: list of additional layers (e.g., Dropout)
        """
        if dataset:
            train_data, val_data = dataset
        else:
            print("🔁 No dataset provided. Using CIFAR-10 dataset as demo...")
            (x_train, y_train), (x_val, y_val) = tf.keras.datasets.cifar10.load_data()
            x_train, x_val = x_train / 255.0, x_val / 255.0
            y_train = tf.keras.utils.to_categorical(y_train, 10)
            y_val = tf.keras.utils.to_categorical(y_val, 10)
            train_data = tf.data.Dataset.from_tensor_slices((x_train, y_train)).shuffle(10000).batch(batch_size)
            val_data = tf.data.Dataset.from_tensor_slices((x_val, y_val)).batch(batch_size)

        print("🚀 Starting training...")

        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Input(shape=(32, 32, 3)))
        model.add(tf.keras.layers.Conv2D(32, (3, 3), activation='relu'))
        model.add(tf.keras.layers.MaxPooling2D((2, 2)))
        model.add(tf.keras.layers.Flatten())
        model.add(tf.keras.layers.Dense(64, activation='relu'))

        if custom_layers:
            for layer in custom_layers:
                model.add(layer)

        model.add(tf.keras.layers.Dense(10, activation='softmax'))

        model.compile(optimizer=optimizer, loss=loss, metrics=['accuracy'])

        checkpoint_path = f"{model_name}.keras"
        checkpoint = tf.keras.callbacks.ModelCheckpoint(checkpoint_path, monitor='val_accuracy', save_best_only=True)

        model.fit(train_data, validation_data=val_data, epochs=epochs, callbacks=[checkpoint])

        self.custom_model = model
        print(f"✅ Training complete. Best model saved as: {checkpoint_path}")

    def Load_Custom_Model(self, path):
        try:
            self.custom_model = tf.keras.models.load_model(path)
            print(f"📂 Model loaded from: {path}")
        except Exception as e:
            print(f"❌ Failed to load model: {e}")

    def Add_Layers(self, base_model, layers):
        """
        Add custom layers to an existing model.

        Parameters:
        - base_model: tf.keras.Sequential or Model to extend.
        - layers: list of tf.keras.layers.Layer to add.
        """
        for layer in layers:
            base_model.add(layer)
        print("✅ Layers added.")
        return base_model
